import json
import os
from cert_schema import normalize_jsonld, extend_preloaded_context

from cert_issuer.config import CONFIG


class JSONLDHandler:
    @staticmethod
    def normalize_to_utf8(certificate_json):
        JSONLDHandler.preload_contexts()
        normalized = normalize_jsonld(certificate_json, detect_unmapped_fields=False)
        return normalized.encode('utf-8')

    @staticmethod
    def preload_contexts():
        if CONFIG.context_urls is None or CONFIG.context_file_paths is None:
            return
        for (url, path) in zip(CONFIG.context_urls, CONFIG.context_file_paths):
            with open(os.path.join(os.getcwd(), path)) as context_file:
                context_data = json.load(context_file)
                extend_preloaded_context(url, context_data)



import unittest

from cert_core import Chain
from pycoin.serialize import b2h

from cert_issuer.merkle_tree_generator import MerkleTreeGenerator
from cert_issuer import helpers
from lds_merkle_proof_2019.merkle_proof_2019 import MerkleProof2019


def get_test_data_generator():
    """
    Returns a generator (1-time iterator) of test data
    :return:
    """
    for num in range(1, 4):
        yield str(num).encode('utf-8')


class TestMerkleTreeGenerator(unittest.TestCase):
    def test_generate(self):
        merkle_tree_generator = MerkleTreeGenerator()
        merkle_tree_generator.populate(get_test_data_generator())
        byte_array = merkle_tree_generator.get_blockchain_data()
        self.assertEqual(b2h(byte_array), '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044')

    def test_proofs_bitcoin_mainnet(self):
        self.do_test_signature(Chain.bitcoin_mainnet, 'bitcoinMainnet', 'BTCOpReturn')

    def test_proofs_bitcoin_testnet(self):
        self.do_test_signature(Chain.bitcoin_testnet, 'bitcoinTestnet', 'BTCOpReturn')

    def test_proofs_bitcoin_regtest(self):
        self.do_test_signature(Chain.bitcoin_regtest, 'bitcoinRegtest', 'BTCOpReturn')

    def test_proofs_mock(self):
        self.do_test_signature(Chain.mockchain, 'mockchain', 'Mock')

    def do_test_signature(self, chain, display_chain, type):
        merkle_tree_generator = MerkleTreeGenerator()
        merkle_tree_generator.populate(get_test_data_generator())
        _ = merkle_tree_generator.get_blockchain_data()
        gen = merkle_tree_generator.get_proof_generator(
            '8087c03e7b7bc9ca7b355de9d9d8165cc5c76307f337f0deb8a204d002c8e582', 'http://example.com', chain)
        p1 = next(gen)
        _ = next(gen)
        p3 = next(gen)

        p1_json_proof = {
            'path': [
                    {'right': 'd4735e3a265e16eee03f59718b9b5d03019c07d8b6c51f90da3a666eec13ab35'},
                    {'right': '4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce'}
                ],
            'merkleRoot': '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044',
            'targetHash': '6b86b273ff34fce19d6b804eff5a3f5747ada4eaa22f1d49c01e52ddb7875b4b',
            'anchors': [
                helpers.tx_to_blink(chain, '8087c03e7b7bc9ca7b355de9d9d8165cc5c76307f337f0deb8a204d002c8e582')
            ]
        }
        mp2019 = MerkleProof2019()
        proof_value = mp2019.encode(p1_json_proof)

        p1_expected = {
                "type": "MerkleProof2019",
                "created": p1['created'],
                "proofValue": proof_value.decode('utf8'),
                "proofPurpose": "assertionMethod",
                "verificationMethod": "http://example.com"
            }

        p3_json_proof = {
            'path': [
                {'left': '4295f72eeb1e3507b8461e240e3b8d18c1e7bd2f1122b11fc9ec40a65894031a'}
            ],
            'merkleRoot': '0932f1d2e98219f7d7452801e2b64ebd9e5c005539db12d9b1ddabe7834d9044',
            'targetHash': '4e07408562bedb8b60ce05c1decfe3ad16b72230967de01f640b7e4729b49fce',
            'anchors': [
                helpers.tx_to_blink(chain, '8087c03e7b7bc9ca7b355de9d9d8165cc5c76307f337f0deb8a204d002c8e582')
            ]
        }
        mp2019 = MerkleProof2019()
        proof_value = mp2019.encode(p3_json_proof)

        p3_expected = {
            "type": "MerkleProof2019",
            "created": p3['created'],
            "proofValue": proof_value.decode('utf8'),
            "proofPurpose": "assertionMethod",
            "verificationMethod": "http://example.com"
        }

        self.assertEqual(p1, p1_expected)
        self.assertEqual(p3, p3_expected)


if __name__ == '__main__':
    unittest.main()



import copy
CHAINED_PROOF_TYPE = 'ChainedProof2021'

class ChainedProof2021:
    type = ''
    verificationMethod = ''
    chainedProofType = ''
    created = ''
    proofPurpose = ''
    previousProof = ''
    proofValue = ''

    def __init__(self, previous_proof, current_proof):
        self.type = CHAINED_PROOF_TYPE
        self.create_proof_object(current_proof)
        self.set_previous_proof(previous_proof)

    def create_proof_object(self, current_proof):
        self.verificationMethod = current_proof['verificationMethod']
        self.chainedProofType = current_proof['type']
        self.created = current_proof['created']
        self.proofPurpose = current_proof['proofPurpose']
        self.proofValue = current_proof['proofValue']

    def set_previous_proof(self, previous_proof):
        if previous_proof['type'] == CHAINED_PROOF_TYPE:
            previous_proof_to_store = copy.deepcopy(previous_proof)
            previous_proof_to_store['type'] = previous_proof['chainedProofType']
            del previous_proof_to_store['chainedProofType']
            del previous_proof_to_store['previousProof']
            self.previousProof = previous_proof_to_store
        else:
            self.previousProof = previous_proof

    def to_json_object(self):
        return self.__dict__



import logging

from jsonschema import validate
from copy import copy


def validate_metadata_structure(metadata):
    if 'schema' in metadata:
        try:
            json_object = copy(metadata)
            del json_object['schema']
            validate(instance=json_object, schema=metadata['schema'])
        except Exception as e:
            print(e)
            raise Exception('Certificate.metadata object does not match its provided schema')
    else:
        logging.warning("""
            The metadata object provided with the certificate does not include a `schema` property.
            Not defining such property will result in errors in the rendering of the metadata property in the UI projects. 
            """)

    if 'displayOrder' not in metadata:
        logging.warning("""
           The metadata object provided with the certificate does not include a `displayOrder` property.
           Not defining such property will result in errors in the rendering of the metadata property in the UI projects. 
           """)
        return
    else:
        verify_display_order_properties(metadata)


def verify_display_order_properties(metadata):
    display_order = metadata['displayOrder']
    checked_groups = []
    for item in display_order:
        path = item.split('.')
        group = path[0]
        if group not in metadata:
            if group not in checked_groups:
                # \033[1m%s\033[0m: display property name in bold
                logging.warning(
                    "`metadata.displayOrder` property references a group named: \033[1m%s\033[0m which does not exist in metadata object.",
                    group
                )
                checked_groups.append(group)
        else:
            property = path[1]
            if property not in metadata[group]:
                logging.warning(
                    "`metadata.displayOrder` property references a property named: \033[1m%s\033[0m which does not exist in group: \033[1m%s\033[0m.",
                    property,
                    group
                )
            else:
                verify_title_is_set(property, group, metadata)

    pass


def verify_title_is_set(property, group, metadata):
    if 'schema' not in metadata:
        return 
    schema = metadata['schema']

    if 'title' not in schema['properties'][group]['properties'][property]:
        logging.warning(
            """No title has been defined for property: \x1b[1m{0}\x1b[0m in group: \x1b[1m{1}\x1b[0m.
            Title should be defined under path `schema.properties.{1}.properties.{0}.title`""".format(property, group)
        )
    pass



from cert_issuer.chained_proof_2021 import ChainedProof2021
from cert_schema import ContextUrls
from cert_issuer.utils import array_intersect

class ProofHandler:
    def __init__(self):
        self.contextUrls = ContextUrls()

    def add_proof(self, certificate_json, merkle_proof):
        if 'proof' in certificate_json:
            if not isinstance(certificate_json['proof'], list):
                # convert proof to list
                initial_proof = certificate_json['proof']
                certificate_json['proof'] = [initial_proof]
            previous_proof = certificate_json['proof'][-1]
            certificate_json['proof'].append(ChainedProof2021(previous_proof, merkle_proof).to_json_object())
            self.update_context_for_chained_proof(certificate_json)
        else:
            certificate_json['proof'] = merkle_proof
            self.update_context_for_single_proof(certificate_json)
        return certificate_json

    def update_context_for_chained_proof(self, certificate_json):
        context = certificate_json['@context']
        if self.contextUrls.merkle_proof_2019() not in context:
            context.append(self.contextUrls.merkle_proof_2019())

        if self.contextUrls.chained_proof_2021() not in context:
            context.append(self.contextUrls.chained_proof_2021())

        if array_intersect(self.contextUrls.v3_all(), context):
            for v3_context in self.contextUrls.v3_all():
                if v3_context in context:
                    index = context.index(v3_context)
                    del context[index]
            context.append(self.contextUrls.v3_1_canonical())

    def update_context_for_single_proof(self, certificate_json):
        context = certificate_json['@context']
        if array_intersect(self.contextUrls.v3_1_all(), context):
            if self.contextUrls.merkle_proof_2019() not in context:
                context.append(self.contextUrls.merkle_proof_2019())



import unittest
from cert_issuer.proof_handler import ProofHandler
from cert_schema import ContextUrls

class TestProofHandler(unittest.TestCase):
    def setUp(self):
        self.handler = ProofHandler()
        self.contextUrls = ContextUrls()

    def test_single_signature(self):
        fixture_certificate_json = {
            '@context': [
                'https://www.w3.org/2018/credentials/v1',
                'https://w3id.org/blockcerts/v3'
            ],
            'kek': 'kek'
        }
        fixture_proof = {
            'type': 'MerkleProof2019',
            'created': '2022-05-05T08:05:14.912828',
            'proofValue': 'zMcm4LfQFUZkWZyLJp1bqtXF8vkZZwp79x7Nvt5BmN2XV4usLLtDoeqiq3et923mcWfXde4a3m4f57yUZcATCbBXV1byb5AXbV8EzT6E8B9JKf3scvxxZCBVePtV4SrhYysAiLNJ9N2R8LgnpJ47wnQHkaTB1AMxrcLEHUTxm4zJTtQqf9orDLf3L4VoLzmST7ZzsDjuX9cw2hZ3Aazhhjy7swG44xfF1PC73SyCv77pDnJ6BSHm3azmbVG6BXv1EPtwF4J1YRqwojBEWk9nDgduACR7b9qNhQ46ND4B5vL8p3LkqTh',
            'proofPurpose': 'assertionMethod',
            'verificationMethod': 'did:ion:EiA_Z6LQILbB2zj_eVrqfQ2xDm4HNqeJUw5Kj2Z7bFOOeQ#key-1'
        }
        output = self.handler.add_proof(fixture_certificate_json, fixture_proof)
        self.assertEqual(output['proof'], fixture_proof)

    def test_single_signature_3_1_update_context_merkle_proof_2019(self):
        fixture_certificate_json = {
            '@context': [
                'https://www.w3.org/2018/credentials/v1',
                'https://w3id.org/blockcerts/v3.1'
            ],
            'kek': 'kek'
        }
        fixture_proof = {
            'type': 'MerkleProof2019',
            'created': '2022-05-05T08:05:14.912828',
            'proofValue': 'zMcm4LfQFUZkWZyLJp1bqtXF8vkZZwp79x7Nvt5BmN2XV4usLLtDoeqiq3et923mcWfXde4a3m4f57yUZcATCbBXV1byb5AXbV8EzT6E8B9JKf3scvxxZCBVePtV4SrhYysAiLNJ9N2R8LgnpJ47wnQHkaTB1AMxrcLEHUTxm4zJTtQqf9orDLf3L4VoLzmST7ZzsDjuX9cw2hZ3Aazhhjy7swG44xfF1PC73SyCv77pDnJ6BSHm3azmbVG6BXv1EPtwF4J1YRqwojBEWk9nDgduACR7b9qNhQ46ND4B5vL8p3LkqTh',
            'proofPurpose': 'assertionMethod',
            'verificationMethod': 'did:ion:EiA_Z6LQILbB2zj_eVrqfQ2xDm4HNqeJUw5Kj2Z7bFOOeQ#key-1'
        }
        output = self.handler.add_proof(fixture_certificate_json, fixture_proof)
        self.assertIn(self.contextUrls.merkle_proof_2019(), output['@context'])

    def test_multiple_two_chained_signature(self):
        fixture_initial_proof = {
          'type': 'Ed25519Signature2020',
          'created': '2022-05-02T16:36:22.933Z',
          'verificationMethod': 'did:key:z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs#z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs',
          'proofPurpose': 'assertionMethod',
          'proofValue': 'zAvFt59599JweBZ4zPP6Ge8LhKgECtBvDRmjG5VQbgEkPCiyMcM9QAPanJgSCs6RRGcKu96qNpfmpe9eTygpFZP6'
        }
        fixture_certificate_json = {
            '@context': [
                'https://www.w3.org/2018/credentials/v1',
                'https://w3id.org/blockcerts/v3'
            ],
            'kek': 'kek',
            'proof': fixture_initial_proof
        }
        fixture_proof = {
            'type': 'MerkleProof2019',
            'created': '2022-05-05T08:05:14.912828',
            'proofValue': 'zMcm4LfQFUZkWZyLJp1bqtXF8vkZZwp79x7Nvt5BmN2XV4usLLtDoeqiq3et923mcWfXde4a3m4f57yUZcATCbBXV1byb5AXbV8EzT6E8B9JKf3scvxxZCBVePtV4SrhYysAiLNJ9N2R8LgnpJ47wnQHkaTB1AMxrcLEHUTxm4zJTtQqf9orDLf3L4VoLzmST7ZzsDjuX9cw2hZ3Aazhhjy7swG44xfF1PC73SyCv77pDnJ6BSHm3azmbVG6BXv1EPtwF4J1YRqwojBEWk9nDgduACR7b9qNhQ46ND4B5vL8p3LkqTh',
            'proofPurpose': 'assertionMethod',
            'verificationMethod': 'did:ion:EiA_Z6LQILbB2zj_eVrqfQ2xDm4HNqeJUw5Kj2Z7bFOOeQ#key-1'
        }
        output = self.handler.add_proof(fixture_certificate_json, fixture_proof)
        self.assertEqual(output['proof'], [
            fixture_initial_proof,
            {
                'type': 'ChainedProof2021',
                'chainedProofType':  'MerkleProof2019',
                'created': '2022-05-05T08:05:14.912828',
                'previousProof': fixture_initial_proof,
                'proofValue': 'zMcm4LfQFUZkWZyLJp1bqtXF8vkZZwp79x7Nvt5BmN2XV4usLLtDoeqiq3et923mcWfXde4a3m4f57yUZcATCbBXV1byb5AXbV8EzT6E8B9JKf3scvxxZCBVePtV4SrhYysAiLNJ9N2R8LgnpJ47wnQHkaTB1AMxrcLEHUTxm4zJTtQqf9orDLf3L4VoLzmST7ZzsDjuX9cw2hZ3Aazhhjy7swG44xfF1PC73SyCv77pDnJ6BSHm3azmbVG6BXv1EPtwF4J1YRqwojBEWk9nDgduACR7b9qNhQ46ND4B5vL8p3LkqTh',
                'proofPurpose': 'assertionMethod',
                'verificationMethod': 'did:ion:EiA_Z6LQILbB2zj_eVrqfQ2xDm4HNqeJUw5Kj2Z7bFOOeQ#key-1'
            }
        ])

    def test_multiple_three_chained_signature(self):
        fixture_initial_proof = {
          'type': 'Ed25519Signature2020',
          'created': '2022-05-02T16:36:22.933Z',
          'verificationMethod': 'did:key:z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs#z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs',
          'proofPurpose': 'assertionMethod',
          'proofValue': 'zAvFt59599JweBZ4zPP6Ge8LhKgECtBvDRmjG5VQbgEkPCiyMcM9QAPanJgSCs6RRGcKu96qNpfmpe9eTygpFZP6'
        }
        fixture_second_proof = {
            'type': 'ChainedProof2021',
            'chainedProofType':  'MerkleProof2019',
            'created': '2022-05-05T08:05:14.912828',
            'previousProof': fixture_initial_proof,
            'proofValue': 'zMcm4LfQFUZkWZyLJp1bqtXF8vkZZwp79x7Nvt5BmN2XV4usLLtDoeqiq3et923mcWfXde4a3m4f57yUZcATCbBXV1byb5AXbV8EzT6E8B9JKf3scvxxZCBVePtV4SrhYysAiLNJ9N2R8LgnpJ47wnQHkaTB1AMxrcLEHUTxm4zJTtQqf9orDLf3L4VoLzmST7ZzsDjuX9cw2hZ3Aazhhjy7swG44xfF1PC73SyCv77pDnJ6BSHm3azmbVG6BXv1EPtwF4J1YRqwojBEWk9nDgduACR7b9qNhQ46ND4B5vL8p3LkqTh',
            'proofPurpose': 'assertionMethod',
            'verificationMethod': 'did:ion:EiA_Z6LQILbB2zj_eVrqfQ2xDm4HNqeJUw5Kj2Z7bFOOeQ#key-1'
        }
        fixture_certificate_json = {
            '@context': [
                'https://www.w3.org/2018/credentials/v1',
                'https://w3id.org/blockcerts/v3'
            ],
            'kek': 'kek',
            'proof': [
                fixture_initial_proof,
                fixture_second_proof
            ]
        }
        fixture_proof = {
            'type': 'MerkleProof2019',
            'created': '2022-05-06T20:31:54',
            'proofValue': 'mockProofValueForUnitTestPurpose',
            'proofPurpose': 'assertionMethod',
            'verificationMethod': 'did:example:ebfeb1f712ebc6f1c276e12ec21#assertion'
        }
        output = self.handler.add_proof(fixture_certificate_json, fixture_proof)
        self.maxDiff = None
        self.assertEqual(output['proof'], [
            fixture_initial_proof,
            fixture_second_proof,
            {
                'type': 'ChainedProof2021',
                'chainedProofType': 'MerkleProof2019',
                'created': '2022-05-06T20:31:54',
                'previousProof': {
                    'type': 'MerkleProof2019',
                    'created': '2022-05-05T08:05:14.912828',
                    'proofValue': 'zMcm4LfQFUZkWZyLJp1bqtXF8vkZZwp79x7Nvt5BmN2XV4usLLtDoeqiq3et923mcWfXde4a3m4f57yUZcATCbBXV1byb5AXbV8EzT6E8B9JKf3scvxxZCBVePtV4SrhYysAiLNJ9N2R8LgnpJ47wnQHkaTB1AMxrcLEHUTxm4zJTtQqf9orDLf3L4VoLzmST7ZzsDjuX9cw2hZ3Aazhhjy7swG44xfF1PC73SyCv77pDnJ6BSHm3azmbVG6BXv1EPtwF4J1YRqwojBEWk9nDgduACR7b9qNhQ46ND4B5vL8p3LkqTh',
                    'proofPurpose': 'assertionMethod',
                    'verificationMethod': 'did:ion:EiA_Z6LQILbB2zj_eVrqfQ2xDm4HNqeJUw5Kj2Z7bFOOeQ#key-1'
                },
                'proofValue': 'mockProofValueForUnitTestPurpose',
                'proofPurpose': 'assertionMethod',
                'verificationMethod': 'did:example:ebfeb1f712ebc6f1c276e12ec21#assertion'
            }
        ])

    def test_adds_merkle_proof_context(self):
        fixture_initial_proof = {
            'type': 'Ed25519Signature2020',
            'created': '2022-05-02T16:36:22.933Z',
            'verificationMethod': 'did:key:z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs#z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs',
            'proofPurpose': 'assertionMethod',
            'proofValue': 'zAvFt59599JweBZ4zPP6Ge8LhKgECtBvDRmjG5VQbgEkPCiyMcM9QAPanJgSCs6RRGcKu96qNpfmpe9eTygpFZP6'
        }
        fixture_certificate_json = {
            '@context': [
                'https://www.w3.org/2018/credentials/v1',
                'https://w3id.org/security/suites/ed25519-2020/v1',
                'https://w3id.org/blockcerts/v3'
            ],
            'kek': 'kek',
            'proof': [
                fixture_initial_proof
            ]
        }
        fixture_proof = {
            'type': 'MerkleProof2019',
            'created': '2022-05-06T20:31:54',
            'proofValue': 'mockProofValueForUnitTestPurpose',
            'proofPurpose': 'assertionMethod',
            'verificationMethod': 'did:example:ebfeb1f712ebc6f1c276e12ec21#assertion'
        }
        output = self.handler.add_proof(fixture_certificate_json, fixture_proof)
        self.assertIn(self.contextUrls.merkle_proof_2019(), output['@context'])

    def test_adds_chained_proof_context(self):
        fixture_initial_proof = {
            'type': 'Ed25519Signature2020',
            'created': '2022-05-02T16:36:22.933Z',
            'verificationMethod': 'did:key:z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs#z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs',
            'proofPurpose': 'assertionMethod',
            'proofValue': 'zAvFt59599JweBZ4zPP6Ge8LhKgECtBvDRmjG5VQbgEkPCiyMcM9QAPanJgSCs6RRGcKu96qNpfmpe9eTygpFZP6'
        }
        fixture_certificate_json = {
            '@context': [
                'https://www.w3.org/2018/credentials/v1',
                'https://w3id.org/security/suites/ed25519-2020/v1',
                'https://w3id.org/blockcerts/v3'
            ],
            'kek': 'kek',
            'proof': [
                fixture_initial_proof
            ]
        }
        fixture_proof = {
            'type': 'MerkleProof2019',
            'created': '2022-05-06T20:31:54',
            'proofValue': 'mockProofValueForUnitTestPurpose',
            'proofPurpose': 'assertionMethod',
            'verificationMethod': 'did:example:ebfeb1f712ebc6f1c276e12ec21#assertion'
        }
        output = self.handler.add_proof(fixture_certificate_json, fixture_proof)
        self.assertIn(self.contextUrls.chained_proof_2021(), output['@context'])

    def test_updates_blockcerts_context_version(self):
        fixture_initial_proof = {
            'type': 'Ed25519Signature2020',
            'created': '2022-05-02T16:36:22.933Z',
            'verificationMethod': 'did:key:z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs#z6MkjHnntGvtLjwfAMHWTAXXGJHhVL3DPtaT9BHmyTjWpjqs',
            'proofPurpose': 'assertionMethod',
            'proofValue': 'zAvFt59599JweBZ4zPP6Ge8LhKgECtBvDRmjG5VQbgEkPCiyMcM9QAPanJgSCs6RRGcKu96qNpfmpe9eTygpFZP6'
        }
        fixture_certificate_json = {
            '@context': [
                'https://www.w3.org/2018/credentials/v1',
                'https://w3id.org/security/suites/ed25519-2020/v1',
                'https://w3id.org/blockcerts/v3'
            ],
            'kek': 'kek',
            'proof': [
                fixture_initial_proof
            ]
        }
        fixture_proof = {
            'type': 'MerkleProof2019',
            'created': '2022-05-06T20:31:54',
            'proofValue': 'mockProofValueForUnitTestPurpose',
            'proofPurpose': 'assertionMethod',
            'verificationMethod': 'did:example:ebfeb1f712ebc6f1c276e12ec21#assertion'
        }
        output = self.handler.add_proof(fixture_certificate_json, fixture_proof)
        self.assertNotIn(self.contextUrls.v3_canonical(), output['@context'])
        self.assertIn(self.contextUrls.v3_1_canonical(), output['@context'])


import re
from urllib.parse import urlparse
from cert_schema import ContextUrls

# TODO: move the v3 checks to cert-schema
def validate_RFC3339_date (date):
    return re.match('^[1-9]\d{3}-\d{2}-\d{2}[Tt\s]\d{2}:\d{2}:\d{2}(?:\.\d{3})?((?:[+-]\d{2}:\d{2})|[Zz])$', date)

def is_valid_url (url):
    try:
        parsed_url = urlparse(url)
    except:
        return False
    return (not (parsed_url.path.__contains__(' ')
       or parsed_url.netloc.__contains__(' '))
       and url.__contains__(':'))

def validate_url (url):
    if not is_valid_url (url):
        raise ValueError('Invalid URL: {}'.format(url))
    pass

def validate_type (certificate_type):
    compulsory_types = ['VerifiableCredential', 'VerifiablePresentation']
    if not isinstance(certificate_type, list):
        raise ValueError('`type` property must be an array')

    contains_compulsory_types = list(set(compulsory_types) & set(certificate_type))
    if len(certificate_type) == 0 or len(contains_compulsory_types) == 0:
        raise ValueError('`type` property must be an array with at least `VerifiableCredential` or `VerifiablePresentation` value')
    pass

def validate_context (context, type):
    ContextUrlsInstance = ContextUrls()
    vc_context_url = ContextUrlsInstance.verifiable_credential()
    blockcerts_valid_context_url = ContextUrlsInstance.v3_all()

    if not isinstance(context, list):
        raise ValueError('`@context` property must be an array')
    if context[0] != vc_context_url:
        raise ValueError('First @context declared must be {}, was given {}'.format(vc_context_url, context[0]))
    if len(type) > 1 and len(context) == 1:
        raise ValueError('A more specific type: {}, was detected, yet no context seems provided for that type'.format(type[1]))
    if context[-1] not in blockcerts_valid_context_url:
        raise ValueError('Last @context declared must be one of valid blockcerts context url, preferably {}, was given {}'.format(blockcerts_valid_context_url[-1]), context[-1])

    pass

def validate_credential_subject (credential_subject):
    pass

def validate_issuer (certificate_issuer):
    if not is_valid_url(certificate_issuer) and not is_valid_url(certificate_issuer['id']):
        raise ValueError('`issuer` property must be a URL string or an object with an `id` property containing a URL string')
    pass

def validate_date_RFC3339_string_format (date, property_name):
    error_message = '`{}` property must be a valid RFC3339 string.'.format(property_name)
    if not isinstance(date, str):
        error_message += ' `{}` value is not a string'.format(date)
        raise ValueError(error_message)

    if not validate_RFC3339_date(date):
        error_message += ' Value received: `{}`'.format(date)
        raise ValueError(error_message)
    pass

def validate_issuance_date (certificate_issuance_date):
    validate_date_RFC3339_string_format(certificate_issuance_date, 'issuanceDate')
    pass

def validate_expiration_date (certificate_expiration_date):
    validate_date_RFC3339_string_format(certificate_expiration_date, 'expirationDate')
    pass

def validate_credential_status (certificate_credential_status):
    try:
        validate_url(certificate_credential_status['id'])
    except KeyError:
        raise ValueError('credentialStatus.id must be defined')
    except ValueError:
        raise ValueError('credentialStatus.id must be a valid URL')

    try:
        isinstance(certificate_credential_status['type'], str)
    except KeyError:
        raise ValueError('credentialStatus.type must be defined')
    except:
        raise ValueError('credentialStatus.type must be a string')
    pass

def verify_credential(certificate_metadata):
    try:
        # if undefined will throw KeyError
        validate_credential_subject(certificate_metadata['credentialSubject'])
    except:
        raise ValueError('`credentialSubject` property must be defined')

    try:
        # if undefined will throw KeyError
        validate_issuer(certificate_metadata['issuer'])
    except KeyError:
        raise ValueError('`issuer` property must be defined')
    except ValueError as err:
        raise ValueError(err)

    try:
        # if undefined will throw KeyError
        validate_issuance_date(certificate_metadata['issuanceDate'])
    except KeyError:
        raise ValueError('`issuanceDate` property must be defined')
    except ValueError as err:
        raise ValueError(err)

    try:
        # if undefined will throw KeyError
        validate_expiration_date(certificate_metadata['expirationDate'])
    except KeyError:
        # optional property
        pass
    except ValueError as err:
        raise ValueError(err)

    try:
        # if undefined will throw KeyError
        validate_credential_status(certificate_metadata['credentialStatus'])
    except KeyError:
        # optional property
        pass
    except ValueError as err:
        raise ValueError(err)

    pass

def verify_presentation (certificate_metadata):
    try:
        for credential in certificate_metadata['verifiableCredential']:
            verify_credential(credential)
    except:
        raise ValueError('A Verifiable Presentation must contain valid verifiableCredential(s)')
    pass